//  Linéo
//  Copyright (C) 2006-2007 Jonathan Buron <nathan@natsimhan.com>
//
//  This file is part of Linéo.
//
//  Linéo is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.

//  Linéo is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with Linéo; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

#include "Musique.h"

// Initialisation du singleton à NULL
Musique *Musique::_Singleton = NULL;

Musique::Musique()
{
    _Music      = NULL;
    _Thread     = NULL;
    _Duration   = 0;
}

Musique::~Musique()
{
    // on attend que le thread se termine tout seul de façon à éviter tout problème
    if(_Thread != NULL) SDL_WaitThread(_Thread,NULL);
    // on libère la musique
    Mix_FreeMusic(_Music);
}

// permet de fournir une instance unique de la classe grâce au singleton
// l'instance est créé si elle n'existe pas encore...
Musique * Musique::getInstance()
{
    if(_Singleton == NULL){
        _Singleton =  new Musique;
    }
    return _Singleton;
}

// détruit l'instance unique de la classe, cette méthode doit être appelée pour détruire la classe
void Musique::kill()
{
    if(_Singleton != NULL){
        // On quitte SDL_mixer
        Mix_CloseAudio();
        delete _Singleton;
        _Singleton = NULL;
    }
}

// initialise la classe
void Musique::init(string file_name)
{
    // on enregistre le nom du fichier
    _FileName = file_name;
    // on initialise SDL_mixer
    if(Mix_OpenAudio( 22050, MIX_DEFAULT_FORMAT, 2, 4096 ) == -1 ){
        fprintf(stderr,
              "\nErreur: Je ne peux pas initialiser le son!\n"
              "L'erreur que SDL a renvoyé était:\n"
              "%s\n\n", SDL_GetError());
        _Music = NULL;
    }else{
        // on charge la musique
        _Music = Mix_LoadMUS(((string)(DATA_PATH)+"son/"+_FileName).c_str());
        if(!_Music){
            fprintf(stderr,
                      "\nErreur: Je ne peux pas charger en mémoire le son suivant \"%s\".\n"
                      "L'erreur que SDL a renvoyé était:\n"
                      "%s\n\n", ((string)(DATA_PATH)+"son/"+_FileName).c_str(), Mix_GetError());
            _Music = NULL;
        }
    }
}

// lance la lecture de la musique
void Musique::play(){
    play(-1);
}
void Musique::play(int loop)
{
    if(_Music == NULL) return;
    //S'il n'y a pas de musique jouée
    if(Mix_PlayingMusic()==0){
        // on met le son a fond
        if(!Mix_SetPanning(MIX_CHANNEL_POST, 255, 255)) {
            fprintf(stderr,
                  "\nErreur: Je ne peux pas mettre le son à fond.\n"
                  "L'erreur que SDL a renvoyé était (Mix_SetPanning) :\n"
                  "%s\n\n", Mix_GetError());
            // on arrête brutalement le son, car on arrive pas à faire de crechendo, et il faut bien l'arrêter...
            stop();
            // et on termine le thread, car ça ne sert à rien de rester...
            return;
        }
        //On lance la musique
        if( Mix_PlayMusic( _Music,loop)==-1){
            fprintf(stderr,
                  "\nErreur: Je ne peux pas jouer le son suivant \"%s\".\n"
                  "L'erreur que SDL a renvoyé était:\n"
                  "%s\n\n", ((string)(DATA_PATH)+"son/"+_FileName).c_str(), Mix_GetError());
        }
    }
}

// arrête la lecture de la musique
void Musique::stop()
{
    if(_Music == NULL) return;
    Mix_HaltMusic();
}

// lance le thread permettant de faire un decrenchendo
// duration est en ms, c'est la durée du decrenchendo
void Musique::decrechendo(Uint32 duration)
{
    // on enregistre la durée pour plus tard
    _Duration = duration;
    // on attend que le thread se termine tout seul de façon à éviter tout problème
    if(_Thread != NULL) SDL_WaitThread(_Thread,NULL);
    //on lance un thread comme ça on ne bloque rien et le decrechendo se fait tout seul dans son coin ;-)
    _Thread = SDL_CreateThread(this->staticDecrechendo, this);
    if ( _Thread == NULL ) {
        fprintf(stderr, "\nImpossible de créer le thread this->staticDecrechendo\n%s\n\n", SDL_GetError());
        return;
    }
}

// réalise un decrechendo sur la musique pendant un temps donné
// l'idée est de diminuer par pas de 1 le niveau sonore et faire un delay entre chaque
// la longueur de ce delay dépend de la durée totale souhaitée divisée par le nombre de pas
// il y a 256 pas (0 à 255)
int Musique::decrechendoThread(void *data)
{
    // on recupere l'instance de la classe VueGame
    Musique *obj = (Musique *)data;
    // on calcul le temps de delay entre chaque palié du niveau de son
    Uint32 delay = (Uint32) ( obj->_Duration / 256 );
    // on boucle  et on descent d'un pas à chaque fois
    for(Uint16 pas=255;pas>0;pas--){
        // on descend d'un palier le son
        if(!Mix_SetPanning(MIX_CHANNEL_POST, pas, pas)) {
            fprintf(stderr,
                  "\nErreur: Je ne peux pas baisser le son.\n"
                  "L'erreur que SDL a renvoyé était (Mix_SetPanning) :\n"
                  "%s\n\n", Mix_GetError());
            // on arrête brutalement le son, car on arrive pas à faire de crenchendo, et il faut bien l'arrêter...
            stop();
            // et on termine le thread, car ça ne sert à rien de rester...
            return -1;
        }
        // on fait une petite pose avant le prochain pas
        SDL_Delay(delay);
    }
    // on s'assure qu'au final el son est bien coupé
    stop();
    // c'est fini on quitte le thread
    return 0;
}
